# -*- coding: utf-8 -*-
"""
Copy to clipboard command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from __future__ import print_function

import datetime
import math
import sys
import time

from clik import args, parser

from safe.app import safe
from safe.clip import clipboard_drivers
from safe.ec import NO_SUCH_ACCOUNT, NO_SUCH_QUESTION, PASSWORD_NOT_SET, \
    VALIDATION_ERROR
from safe.model import Account, Password


DEFAULT_TIME = 5
JUNK = 'x'
UPDATE_INTERVAL = 0.1


class Countdown(object):
    """Countdown timer on stdout."""

    def __init__(self, format):
        """Format should have a single ``%s``, where time left is inserted."""
        self.format = format
        self._n = 0

    def _print(self, content):
        """Clear current line, replace it with ``content``."""
        out = '\r%s\r%s' % (' ' * self._n, content)
        self._n = len(content)
        sys.stdout.write(out)
        sys.stdout.flush()

    def update(self, time_left):
        """Update stdout with current ``time_left``."""
        self._print(self.format % time_left)

    def end(self):
        """Finish the countdown (by clearing the current line)."""
        self._print('')


@safe(alias='pb')
def copy():
    """Copy secret to clipboard temporarily."""
    parser.add_argument(
        'name',
        nargs=1,
        help='name of the account for which to copy secret',
    )
    parser.add_argument(
        '-q',
        '--question',
        help='copy answer to security question with given identifier to '
             'clipboard instead of the account password',
    )
    parser.add_argument(
        '-t',
        '--time',
        default=DEFAULT_TIME,
        help='amount of time to keep secret on clipboard (default: '
             '%(default)s)',
        type=float,
    )
    clipboard_drivers.configure_parser(parser)

    yield

    if args.time < 1:
        print('error: -t/--time must be one or greater', file=sys.stderr)
        yield VALIDATION_ERROR

    account = Account.for_slug(args.name[0])
    if account is None:
        print('error: no account named', args.name[0], file=sys.stderr)
        yield NO_SUCH_ACCOUNT

    if args.question:
        question = account.question_query\
                          .filter_by(identifier=args.question)\
                          .first()
        if question is None:
            fmt = 'error: no question with identifier "%s" associated with ' \
                  'account "%s"'
            print(fmt % (args.question, account.name), file=sys.stderr)
            yield NO_SUCH_QUESTION
        value = question.answer
    else:
        password = account.password_query\
                          .order_by(Password.changed.desc())\
                          .limit(1)\
                          .first()
        if password is None:
            msg = 'error: no password set for account "%s"' % account.name
            print(msg, file=sys.stderr)
            yield PASSWORD_NOT_SET
        value = password.value

    now = datetime.datetime.today()
    end = now + datetime.timedelta(seconds=args.time)
    countdown = Countdown('secret on clipboard for %ss...')
    clipboard = clipboard_drivers.driver_for_args(args)
    clipboard.put(value)
    while now < end:
        time.sleep(UPDATE_INTERVAL)
        now = datetime.datetime.today()
        countdown.update(math.ceil((end - now).seconds) + 1)
    countdown.end()
    clipboard.put(JUNK)



# -*- coding: utf-8 -*-
"""
List command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from __future__ import print_function

import sys

import sqlalchemy
from clik import args, parser

from safe.app import safe
from safe.cmd.list.account import list_accounts
from safe.cmd.list.policy import list_policies
from safe.ec import NO_SUCH_ACCOUNT, NO_SUCH_POLICY
from safe.model import Account, Alias, Policy


@safe(name='list', alias='ls')
def list_():
    """List objects in the database."""
    parser.add_argument(
        'name',
        nargs='*',
        help='object name(s) to list',
    )
    parser.add_argument(
        '-p',
        '--policy',
        action='store_true',
        default=False,
        help='list policy object(s) instead of account(s)',
    )
    parser.add_argument(
        '-s',
        '--strict',
        action='store_true',
        default=False,
        help='match name/alias strictly',
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        dest='verbosity',
        help='list more detail (may be supplied multiple times)',
    )

    yield

    if args.policy:
        if Policy.query.count() == 0:
            print('no policies in the database')
            print('run "safe new policy -h" for help on creating policies')
            yield NO_SUCH_POLICY

        if args.strict and args.name:
            where = Policy.name.in_(args.name)
        else:
            clauses = [Policy.name.ilike('%%%s%%' % n) for n in args.name]
            where = sqlalchemy.or_(*clauses)

        policies = Policy.query.filter(where)
        if policies.count() < 1:
            print('error: no policy matches the query', file=sys.stderr)
            yield NO_SUCH_POLICY

        list_policies(policies, args.verbosity)
    else:
        if Account.query.count() == 0:
            print('no accounts in the database')
            print('run "safe new account -h" for help on creating accounts')
            yield NO_SUCH_ACCOUNT

        if args.strict and args.name:
            where = Account.name.in_(args.name) | Alias.value.in_(args.name)
        else:
            where = sqlalchemy.or_()
            for name in args.name:
                like = '%%%s%%' % name
                where |= Account.name.ilike(like) | Alias.value.ilike(like)

        accounts = Account.query.outerjoin(Alias).filter(where)
        if accounts.count() < 1:
            print('error: no account matches the query', file=sys.stderr)
            yield NO_SUCH_ACCOUNT

        list_accounts(accounts, args.verbosity)



# -*- coding: utf-8 -*-
"""
Drop object command group.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from safe.app import safe


@safe(alias='rm')
def drop():
    """Drop objects from the database."""
    yield



# -*- coding: utf-8 -*-
"""
Data model for safe.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
import datetime
import re

from clik import g

from safe.db import ORM
from safe.sgen import generate


#: ORM wrapper instance.
#:
#: :type: :class:`safe.db.ORM`
orm = ORM()

#: Represents the ``cascade`` arguments necessary to effect deletion.
#:
#: :type: :class:`str`
CASCADE_DELETE = 'save-update, merge, delete'

#: Maximum length of a slug.
#:
#: :type: :class:`int`
SLUG_LENGTH = 20

#: Regular expression matching a valid slug.
#:
#: :type: :func:`re.compile`
SLUG_RE = re.compile(r'^[a-zA-Z0-9_/-]{1,%i}$' % SLUG_LENGTH)

#: User-level validation error message when a value does not match
#: :data:`SLUG_RE`.
#:
#: :type: :class:`str`
SLUG_VALIDATION_ERROR_MESSAGE = 'value must be 1-%i characters and contain ' \
                                'only letters, numbers, underscores, ' \
                                'forward slashes, and hyphens' % SLUG_LENGTH


def validate_slug(value):
    """
    Assert that ``value`` matches :data:`SLUG_RE`, return ``value`` if so.

    :param str value: Value to validate
    :raise: :exc:`AssertionError` if ``value`` does not match :data:`SLUG_RE`
    :return: Original value if valid
    :rtype: :class:`str`
    """
    assert SLUG_RE.search(value)
    return value


class Account(orm.Model):
    """Account model around which the whole application revolves."""

    __tablename__ = 'account'

    #: Primary key.
    #:
    #: :type: :class:`int`
    id = orm.Column(orm.Integer, primary_key=True)

    #: Query for :class:`Alias` instances.
    #:
    #: :type: :func:`sqlalchemy.orm.relationship` to :class:`Alias`
    alias_query = orm.relationship('Alias', lazy='dynamic')

    #: Eagerly loaded list of :class:`Alias` instances.
    #:
    #: :type: :class:`list` of :class:`Alias` instances
    aliases = orm.relationship('Alias', cascade=CASCADE_DELETE)

    #: Query for :class:`Code` instances.
    #:
    #: :type: :func:`sqlalchemy.orm.relationship` to :class:`Code`
    code_query = orm.relationship('Code', lazy='dynamic')

    #: Eagerly loaded list of :class:`Code` instances.
    #:
    #: :type: :class:`list` of :class:`Code` instances
    codes = orm.relationship('Code', cascade=CASCADE_DELETE)

    #: *(optional)* Short description for the account.
    #:
    #: :type: :class:`str` or ``None``
    description = orm.Column(orm.Text)

    #: *(optional)* Email associated with the account.
    #:
    #: :type: :class:`str` or ``None``
    email = orm.Column(orm.Text)

    #: Canonical name for the account.
    #:
    #: :type: :class:`str` matching :data:`SLUG_RE`
    name = orm.Column(orm.String(SLUG_LENGTH), nullable=False, unique=True)

    #: Query for :class:`Password` instances.
    #:
    #: :type: :func:`sqlalchemy.orm.relationship` to :class:`Password`
    password_query = orm.relationship('Password', lazy='dynamic')

    #: Eagerly loaded list of :class:`Password` instances.
    #:
    #: :type: :class:`list` of :class:`Password` instances
    passwords = orm.relationship('Password', cascade=CASCADE_DELETE)

    #: *(optional)* Foreign key of :class:`Policy` for passwords for
    #: this account.
    #:
    #: :type: :class:`int` or ``None``
    password_policy_id = orm.Column(orm.Integer, orm.ForeignKey('policy.id'))

    #: *(optional)* :class:`Policy` for passwords for this account.
    #:
    #: :type: :class:`Policy` or ``None``
    password_policy = orm.relationship(
        'Policy', foreign_keys=[password_policy_id])

    #: *(optional)* Foreign key of :class:`Policy` for security
    #: question answers for this account.
    #:
    #: :type: :class:`int` or ``None``
    question_policy_id = orm.Column(orm.Integer, orm.ForeignKey('policy.id'))

    #: *(optional)* :class:`Policy` for security question answers for
    #: this account.
    #:
    #: :type: :class:`Policy` or ``None``
    question_policy = orm.relationship(
        'Policy', foreign_keys=[question_policy_id])

    #: Query for :class:`Question` instances.
    #:
    #: :type: :func:`sqlalchemy.orm.relationship` to :class:`Question`
    question_query = orm.relationship('Question', lazy='dynamic')

    #: Eagerly loaded list of :class:`Question` instances.
    #:
    #: :type: :class:`list` of :class:`Question` instances
    questions = orm.relationship('Question', cascade=CASCADE_DELETE)

    #: *(optional)* Username associated with the account.
    #:
    #: :type: :class:`str` or ``None``
    username = orm.Column(orm.Text)

    @staticmethod
    def _filter_slug(query, slug):
        return query.outerjoin(Account.aliases)\
                    .filter((Account.name == slug) | (Alias.value == slug))

    @classmethod
    def id_for_slug(cls, slug):
        """
        Return ID for account with name or alias ``slug``.

        :return: ID if account exists for ``slug``, else ``None``
        :rtype: :class:`int` or ``None``
        """
        return cls._filter_slug(g.db.query(Account.id), slug).scalar()

    @classmethod
    def for_slug(cls, slug):
        """
        Return instance for account with name or alias ``slug``.

        :return: Account instance if account exists for ``slug``, else ``None``
        :rtype: :class:`Account` or ``None``
        """
        return cls._filter_slug(g.db.query(Account), slug).first()

    @orm.validates('name')
    def validate_name(self, _, name):
        """Validate that ``name`` matches :data:`SLUG_RE`."""
        return validate_slug(name)


class Alias(orm.Model):
    """Alias associated with an account."""

    __tablename__ = 'alias'

    #: Primary key.
    #:
    #: :type: :class:`int`
    id = orm.Column(orm.Integer, primary_key=True)

    #: Foreign key of :class:`Account` with which this alias is
    #: associated.
    #:
    #: :type: :class:`int`
    account_id = orm.Column(
        orm.Integer, orm.ForeignKey('account.id'), nullable=False)

    #: :class:`Account` instance with which this alias is associated.
    #:
    #: :type: :class:`Account`
    account = orm.relationship('Account')

    #: The alias.
    #:
    #: :type: :class:`str` matching :data:`SLUG_RE`
    value = orm.Column(orm.String(SLUG_LENGTH), nullable=False, unique=True)

    @orm.validates('value')
    def validate_value(self, _, value):
        """Validate that ``value`` matches :data:`SLUG_RE`."""
        return validate_slug(value)


class Code(orm.Model):
    """Backup code associated with an account."""

    __tablename__ = 'code'
    __table_args__ = (orm.UniqueConstraint(
        'account_id', 'value', name='_account_id_value_uc'),)

    #: Primary key.
    #:
    #: :type: :class:`int`
    id = orm.Column(orm.Integer, primary_key=True)

    #: Foreign key of :class:`Account` with which this code is
    #: associated.
    #:
    #: :type: :class:`int`
    account_id = orm.Column(
        orm.Integer, orm.ForeignKey('account.id'), nullable=False)

    #: :class:`Account` instance with which this code is associated.
    #:
    #: :type: :class:`Account`
    account = orm.relationship('Account')

    #: Indicates whether the code has been used.
    #:
    #: :type: :class:`str`
    used = orm.Column(orm.Boolean, default=False, nullable=False)

    #: The code.
    #:
    #: :type: :class:`str`
    value = orm.Column(orm.Text, nullable=False)


class Password(orm.Model):
    """Password associated with an account."""

    __tablename__ = 'password'

    #: Primary key.
    #:
    #: :type: :class:`int`
    id = orm.Column(orm.Integer, primary_key=True)

    #: Foreign key of :class:`Account` with which this password is
    #: associated.
    #:
    #: :type: :class:`int`
    account_id = orm.Column(
        orm.Integer, orm.ForeignKey('account.id'), nullable=False)

    #: :class:`Account` instance with which this password is
    #: associated.
    #:
    #: :type: :class:`Account`
    account = orm.relationship('Account')

    #: Datetime at which this password was set.
    #:
    #: :type: :class:`datetime.datetime`
    changed = orm.Column(
        orm.DateTime, default=datetime.datetime.today, nullable=False)

    #: The password.
    #:
    #: :type: :class:`str`
    value = orm.Column(orm.Text, nullable=False)


class Policy(orm.Model):
    """Policy specifying how secrets should be generated/rotated."""

    #: Default value for :attr:`frequency`.
    #:
    #: :type: :class:`int`
    DEFAULT_FREQUENCY = 0

    #: Default value for :attr:`generator`.
    #:
    #: :type: :class:`str`
    DEFAULT_GENERATOR = 'default'

    #: Default value for :attr:`length`.
    #:
    #: :type: :class:`int`
    DEFAULT_LENGTH = 24

    __tablename__ = 'policy'

    #: Primary key.
    #:
    #: :type: :class:`int`
    id = orm.Column(orm.Integer, primary_key=True)

    #: *(optional)* Short description for the policy.
    #:
    #: :type: :class:`str` or ``None``
    description = orm.Column(orm.Text)

    #: Characters that may not be used in generated secrets.
    #:
    #: :type: :class:`str`
    disallowed_characters = orm.Column(orm.Text, default='', nullable=False)

    #: Frequency (in days) at which secret should be rotated. 0 means
    #: never.
    #:
    #: :type: :class:`int`
    frequency = orm.Column(
        orm.Integer, default=DEFAULT_FREQUENCY, nullable=False)

    #: Name for the secret generator.
    #:
    #: :type: :class:`str` name of generator in :mod:`safe.generator`
    generator = orm.Column(
        orm.String(79), default=DEFAULT_GENERATOR, nullable=False)

    #: Length of generated secrets.
    #:
    #: :type: :class:`int`
    length = orm.Column(orm.Integer, default=DEFAULT_LENGTH, nullable=False)

    #: Name for the policy.
    #:
    #: :type: :class:`str` matching :data:`SLUG_RE`
    name = orm.Column(orm.String(SLUG_LENGTH), nullable=False, unique=True)

    @classmethod
    def id_for_name(cls, name):
        """
        Return ID for policy named ``name``.

        :return: ID if policy exists, else ``None``
        :rtype: :class:`int` or ``None``
        """
        return g.db.query(Policy.id).filter(Policy.name == name).scalar()

    @classmethod
    def for_name(cls, name):
        """
        Return instance for policy named ``name``.

        :return: Policy instance if policy exists, else ``None``
        :rtype: :class:`Policy` or ``None``
        """
        return g.db.query(Policy).filter(Policy.name == name).first()

    def generate_secret(self):
        """
        Return a secret meeting the parameters of this policy.

        :return: Randomly generated secret
        :rtype: :class:`str`
        """
        generator = generate[self.generator]
        return generator(self.length, self.disallowed_characters)

    @orm.validates('frequency')
    def validate_frequency(self, _, frequency):
        """Validate that ``frequency`` is zero or greater."""
        assert frequency > -1
        return frequency

    @orm.validates('generator')
    def validate_generator(self, _, generator):
        """Validate that ``generator`` is a valid generator name."""
        assert generator in generate
        return generator

    @orm.validates('length')
    def validate_length(self, _, length):
        """Validate that ``length`` is one or greater."""
        assert length > 0
        return length

    @orm.validates('name')
    def validate_name(self, _, name):
        """Validate that ``name`` matches :data:`SLUG_RE`."""
        return validate_slug(name)


class Question(orm.Model):
    """Security question and answer associated with an account."""

    __tablename__ = 'question'
    __table_args__ = (orm.UniqueConstraint(
        'account_id', 'identifier', name='_account_id_identifier_uc'),)

    #: Primary key.
    #:
    #: :type: :class:`int`
    id = orm.Column(orm.Integer, primary_key=True)

    #: Foreign key of :class:`Account` with which this security
    #: question is associated.
    #:
    #: :type: :class:`int`
    account_id = orm.Column(
        orm.Integer, orm.ForeignKey('account.id'), nullable=False)

    #: :class:`Account` instance with which this security question is
    #: associated.
    #:
    #: :type: :class:`Account`
    account = orm.relationship('Account')

    #: Answer to the security question.
    #:
    #: :type: :class:`str`
    answer = orm.Column(orm.Text, nullable=False)

    #: Short, human-friendly identifier for the question.
    #:
    #: :type: :class:`str` matching :data:`SLUG_RE`
    identifier = orm.Column(orm.String(SLUG_LENGTH), nullable=False)

    #: The security question.
    #:
    #: :type: :class:`str`
    question = orm.Column(orm.Text, nullable=False)

    @orm.validates('identifier')
    def validate_identifier(self, _, identifier):
        """Validate that ``identifier`` matches :data:`SLUG_RE`."""
        return validate_slug(identifier)



# -*- coding: utf-8 -*-
"""
Exit codes for the application.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""

#: User canceled the operation.
CANCELED = 10

#: Generic validation error.
VALIDATION_ERROR = 20

#: No account with given name.
NO_SUCH_ACCOUNT = 21

#: No policy with given name.
NO_SUCH_POLICY = 22

#: No policy with given identifier.
NO_SUCH_QUESTION = 23

#: GPG executable not found on system.
MISSING_GPG = 30

#: Decryption of GPG file failed.
DECRYPTION_FAILED = 31

#: Encryption of plaintext file failed.
ENCRYPTION_FAILED = 32

#: File is required, but missing.
MISSING_FILE = 40

#: File to be created already exists.
FILE_EXISTS = 41

#: Command requires file argument, which was not supplied.
FILE_ARGUMENT_REQUIRED = 42

#: Failed to securely delete a plaintext file with sensitive data.
#: (File deleted, but not in a secure manner.)
SRM_FAILED = 100

#: No password has been set for the account.
PASSWORD_NOT_SET = 101



# -*- coding: utf-8 -*-
"""
Subpackage containing the commands for the application.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
import safe.cmd.copy
import safe.cmd.drop.account
import safe.cmd.drop.policy
import safe.cmd.gen.bare
import safe.cmd.gen.per_policy
import safe.cmd.init
import safe.cmd.ipy
import safe.cmd.list
import safe.cmd.new.account
import safe.cmd.new.policy
import safe.cmd.shell
import safe.cmd.update.account
import safe.cmd.update.policy  # noqa: F401



# -*- coding: utf-8 -*-
"""
Account listing helpers.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from safe.cmd.list.policy import get_policy_specs
from safe.cmd.list.util import NONE, NOT_SET, nullable, print_detail, \
    print_table, sorted_by_name
from safe.model import Password, Question


#: :meth:`datetime.datetime.strftime` format in which dates will be
#: printed.
#:
#: :type: :class:`str`
DATE_FORMAT = '%Y-%m-%d'

#: "Template" for question subtrees for prolix output.
#:
#: :type: :class:`str`
prolix_question_fmt = """
    ┃  ┏─┄┈ id → %(identifier)s
    %(box_1)s──╋──┄┈ q → %(question)s
    %(box_2)s  ┖──┄┈ a → %(answer)s
""".strip()


def get_common_details(account):
    """
    Return details common to both verbosity levels.

    :param safe.model.Account account: Account for which to get details
    :return: Sequence of 2-tuples with details for the UI
    :rtype: :func:`tuple` of 2-tuples ``(field name, info)`` where both
            items are strings
    """
    aliases = NONE
    if account.aliases:
        aliases = ', '.join([alias.value for alias in account.aliases])

    password = NOT_SET
    password_count = account.password_query.count()
    if password_count > 0:
        if password_count == 1:
            changed = account.password_query.all()[0].changed
            password = 'set on %s' % changed.strftime(DATE_FORMAT)
        else:
            fmt = 'changed %i times (first: %s) (last: %s)'

            def date(order_by):
                pw = account.password_query.order_by(order_by).limit(1).first()
                return pw.changed.strftime(DATE_FORMAT)

            first = date(Password.changed)
            last = date(Password.changed.desc())
            password = fmt % (password_count, first, last)

    return (
        ('description', nullable(account.description)),
        ('aliases', aliases),
        ('username', nullable(account.username)),
        ('email', nullable(account.email)),
        ('password', password),
    )


def get_verbose_details(account):
    """
    Return details for verbose output.

    :param safe.model.Account account: Account for which to get details
    :return: Sequence of 2-tuples with details for the UI
    :rtype: :func:`tuple` of 2-tuples ``(field name, info)`` where both
            items are strings
    """
    questions = 'no questions'
    questions_count = account.question_query.count()
    if questions_count > 0:
        questions = '%i question' % questions_count
        if questions_count > 1:
            questions += 's'

    codes = 'no backup codes'
    codes_count = account.code_query.count()
    if codes_count > 0:
        codes = '%i backup code' % codes_count
        if codes_count > 1:
            codes += 's'
        used_codes = 'none used'
        used_codes_count = account.code_query.filter_by(used=True).count()
        if used_codes_count > 0:
            used_codes = '%i used' % used_codes_count
        codes += ' (%s)' % used_codes

    security = '%s, %s' % (questions, codes)

    fmt = 'password is %s, question is %s'
    password_policy = '<not set>'
    if account.password_policy:
        password_policy = '"%s"' % account.password_policy.name
    question_policy = '<not set>'
    if account.question_policy:
        question_policy = '"%s"' % account.question_policy.name
    policies = fmt % (password_policy, question_policy)

    return (('security', security), ('policies', policies))


def get_prolix_details(account):
    """
    Return details for prolix output.

    :param safe.model.Account account: Account for which to get details
    :return: Sequence of 2-tuples with details for the UI
    :rtype: :func:`tuple` of 2-tuples ``(field name, info)`` where both
            items are strings
    """
    active_codes = NONE
    active_codes_query = account.code_query.filter_by(used=False)
    if active_codes_query.count() > 0:
        codes = sorted([code.value for code in active_codes_query.all()])
        active_codes = ', '.join(codes)

    used_codes = NONE
    used_codes_query = account.code_query.filter_by(used=True)
    if used_codes_query.count() > 0:
        codes = sorted([code.value for code in used_codes_query.all()])
        used_codes = ', '.join(codes)

    password_policy = 'policy is %s' % NOT_SET
    if account.password_policy:
        p = account.password_policy
        password_policy = '%s (%s)' % (p.name, get_policy_specs(p))

    question_policy = 'policy is %s' % NOT_SET
    if account.question_policy:
        p = account.question_policy
        question_policy = '%s (%s)' % (p.name, get_policy_specs(p))

    question_count = account.question_query.count()

    return (
        ('active codes', active_codes),
        ('used codes', used_codes),
        ('passwords', password_policy),
        ('answers', question_policy),
        ('questions', str(question_count)),
    )


def print_verbosely(account):
    """
    Print verbose information for ``account``.

    :param safe.model.Account account: Account for which to print verbose info
    """
    common_details = get_common_details(account)
    extra_details = get_verbose_details(account)
    print_detail(account.name, common_details + extra_details)


def print_prolixly(account):
    """
    Print prolix information for ``account``.

    :param safe.model.Account account: Account for which to print prolix info
    """
    common_details = get_common_details(account)
    extra_details = get_prolix_details(account)
    hack = None
    if account.question_query.count() > 0:
        hack = (1, '┰')
    print_detail(account.name, common_details + extra_details, hack=hack)
    print_questions(account)


def print_questions(account):
    """
    Print question subtree for ``account``.

    :param safe.model.Account account: Account for which to print question
                                       subtree
    """
    i, n = 0, account.question_query.count()
    for question in account.question_query.order_by(Question.identifier).all():
        i += 1
        box_1 = '┠'
        box_2 = '┃'
        if i == n:
            box_1 = '┖'
            box_2 = ' '
        print('    %s' % prolix_question_fmt % dict(
            answer=question.answer,
            box_1=box_1,
            box_2=box_2,
            identifier=question.identifier,
            question=question.question,
        ))


def list_accounts(accounts, verbosity):
    """
    Print account information in ``verbosity`` level of detail.

    :param accounts: SQLAlchemy query containing accounts to be printed
    :type accounts: :class:`sqlalchemy.orm.query.Query`
    :param int verbosity: Must be at least 0, anything past 2 is ignored,
                          higher means more information
    """
    print()
    if verbosity < 1:
        rows = []
        for a in sorted_by_name(accounts):
            rows.append((a.name, nullable(a.username), nullable(a.email)))
        print_table(('NAME', 'USERNAME', 'EMAIL'), rows)
    elif verbosity < 2:
        for account in sorted_by_name(accounts):
            print_verbosely(account)
    else:
        for account in sorted_by_name(accounts):
            print_prolixly(account)
    print()



# -*- coding: utf-8 -*-
"""
Secret generation utility command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from safe.app import safe


@safe
def gen():
    """Generate secrets to stdout."""
    yield



# -*- coding: utf-8 -*-
"""
Database helpers and ORM shim.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
import contextlib

import sqlalchemy
import sqlalchemy.ext.declarative
import sqlalchemy.orm
import sqlalchemy.types


def open_database(path):
    """
    Return SQLAlchemy session for SQLite database at ``path``.

    :param str path: Path to SQLite file
    :return: SQLAlchemy session
    :rtype: :class:`sqlalchemy.orm.session.Session`
    """
    engine = sqlalchemy.create_engine('sqlite:///%s' % path)
    return sqlalchemy.orm.sessionmaker(bind=engine)()


class ORM(object):
    """Unified wrapper for SQLAlchemy's ORM API."""

    def __init__(self):
        """Initialize the wrapper."""
        for module in (sqlalchemy, sqlalchemy.orm):
            for attr in module.__all__:
                if not hasattr(self, attr):
                    setattr(self, attr, getattr(module, attr))
        self.Model = sqlalchemy.ext.declarative.declarative_base()

    @contextlib.contextmanager
    def bind(self, session):
        """
        Context manager to bind the ORM instance to a SQLAlchemy session.

        This adds a ``query`` convenience property to each model class,
        a la flask-sqlalchemy_.

        .. _flask-sqlalchemy: http://flask-sqlalchemy.pocoo.org/

        :param session: SQLAlchemy session to which to bind
        :type session: :class:`sqlalchemy.orm.session.Session`
        """
        class QueryProperty(object):
            def __init__(self, session):
                self._session = session

            def __get__(self, _, type):
                mapper = sqlalchemy.orm.class_mapper(type)
                return sqlalchemy.orm.Query(mapper, session=self._session)

        self.Model.query = QueryProperty(session)
        try:
            yield
        finally:
            del self.Model.query



# -*- coding: utf-8 -*-
"""
Policy listing helpers.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from safe.cmd.list.util import NONE, nullable, print_detail, print_table, \
    sorted_by_name


def get_policy_specs(policy):
    """
    Return short summary of secret generation params for ``policy``.

    :param safe.model.Policy policy: Policy for which to get specs
    :return: Human-friendly description of the policy
    :rtype: :class:`str`
    """
    if policy.frequency > 0:
        if policy.frequency == 1:
            change = 'changed every day'
        else:
            change = 'changed every %i days' % policy.frequency
    else:
        change = 'never changed'
    fmt = '%i chars from %s, %s'
    return fmt % (policy.length, policy.generator, change)


def list_policies(policies, verbosity):
    """
    Print policy information in ``verbosity`` level of detail.

    :param policies: SQLAlchemy query containing policies to be printed
    :type policies: :class:`sqlalchemy.orm.query.Query`
    :param int verbosity: Must be at least 0, anything past 1 is ignored,
                          higher means more information
    """
    print()
    if verbosity < 1:
        rows = []
        for p in sorted_by_name(policies):
            rows.append((p.name, p.generator, p.length, p.frequency))
        print_table(('NAME', 'GEN', 'LEN', 'FREQ'), rows)
    else:
        for policy in sorted_by_name(policies):
            chars = NONE
            if policy.disallowed_characters:
                chars = ''.join(sorted(policy.disallowed_characters))
            print_detail(
                policy.name, (
                    ('description', nullable(policy.description)),
                    ('specs', get_policy_specs(policy)),
                    ('∅ chars', chars),
                ),
            )
    print()



# -*- coding: utf-8 -*-
"""
Database initialization command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from __future__ import print_function

import os
import subprocess
import sys

import sqlalchemy
from clik import args, parser
from clik_shell import exclude_from_shell

from safe.app import ignore_file_argument, safe
from safe.ec import ENCRYPTION_FAILED, FILE_EXISTS
from safe.gpg import get_gpg_executable
from safe.model import orm
from safe.util import expand_path, temporary_directory


@exclude_from_shell
@safe
def init():
    """Create and initialize the database."""
    ignore_file_argument()

    parser.add_argument(
        'file',
        help='database file',
        nargs=1,
    )

    mode_group = parser.add_mutually_exclusive_group(required=True)
    mode_group.add_argument(
        '-k',
        '--key',
        default=None,
        help='use key-based (asymmetric) encryption (specify the public key '
             'id to encrypt to as the value for this argument)',
        metavar='KEYID',
    )
    mode_group.add_argument(
        '-p',
        '--password',
        action='store_true',
        default=False,
        help='use password-based (symmetric) encryption',
    )

    yield

    path = expand_path(args.file[0])
    if os.path.exists(path):
        print('error: database file already exists:', path, file=sys.stderr)
        yield FILE_EXISTS

    with temporary_directory() as tmp:
        tmp_path = os.path.join(tmp, 'db')

        engine = sqlalchemy.create_engine('sqlite:///%s' % tmp_path)
        metadata = orm.Model.metadata
        metadata.create_all(bind=engine, tables=metadata.tables.values())

        command = (
            get_gpg_executable(),
            '--armor',
            '--cipher-algo', args.cipher,
            '--output', path,
            '--quiet',
        )
        if args.password:
            command += ('--symmetric',)
        else:
            command += ('--recipient', args.key, '--encrypt')
        command += (tmp_path,)

        process = subprocess.Popen(command)
        process.wait()
        if process.returncode:
            yield ENCRYPTION_FAILED



# -*- coding: utf-8 -*-
"""
Update policy command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from __future__ import print_function

import sys

from clik import args, g, parser

from safe.cmd.update import update
from safe.ec import NO_SUCH_POLICY, VALIDATION_ERROR
from safe.form.policy import UpdatePolicyForm
from safe.model import Policy


@update(alias='p')
def policy():
    """Update a policy and/or its associated data."""
    parser.add_argument(
        'policy',
        help='name of policy to update',
        nargs=1,
    )

    form = UpdatePolicyForm()
    form.configure_parser()

    yield

    policy = Policy.for_name(args.policy[0])
    if policy is None:
        print('error: no policy with name:', args.policy[0])
        yield NO_SUCH_POLICY

    if not form.bind_and_validate(policy):
        msg = 'error: there were validation error(s) with input value(s)'
        print(msg, file=sys.stderr)
        form.print_errors()
        yield VALIDATION_ERROR

    form.update_policy()
    g.commit_and_save()



# -*- coding: utf-8 -*-
"""
Secure file deletion utility.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
import os

from safe.util import get_executable, Subprocess


#: Value to pass to ``--iterations`` argument of ``shred``.
#:
#: :type: :class:`int`:
SHRED_ITERATIONS = 35


class SecureDeleteError(Exception):
    """Raised when there is a problem securely deleting a file."""

    def __init__(self, message, stdout, stderr):
        """
        Instantiate the error.

        :param str message: Short message describing the error
        :param stdout: Standard output related to the error
        :type stdout: :class:`str` or ``None``
        :param stderr: Standard error related to the error
        :type stderr: :class:`str` or ``None``
        """
        super(SecureDeleteError, self).__init__(message)

        #: Short message describing the error.
        #:
        #: :type: :class:`str`
        self.message = message

        #: Standard out associated with the error.
        #:
        #: :type: :class:`str` or ``None``
        self.stdout = stdout

        #: Standard error associated with the error.
        #:
        #: :type: :class:`str` or ``None``
        self.stderr = stderr


def secure_delete(path):
    """
    Securely delete file at ``path``.

    :param str path: Path of file to delete
    :raise: :exc:`SecureDeleteError` if there are no secure deletion utilities
            found on the machine, or if the secure deletion utility returns
            a non-zero exit code
    :rtype: ``None``
    """
    srm = get_executable('srm')
    if srm is not None:
        process = Subprocess((srm, path))
        stdout, stderr = process.communicate()
        if process.returncode:
            msg = 'srm returned non-zero exit code: %s' % process.returncode
            raise SecureDeleteError(msg, stdout, stderr)
        return

    shred = get_executable('shred')
    if shred is not None:
        cmd = (shred, '--iterations', str(SHRED_ITERATIONS), path)
        process = Subprocess(cmd)
        stdout, stderr = process.communicate()
        if process.returncode:
            msg = 'shred returned non-zero exit code: %s' % process.returncode
            raise SecureDeleteError(msg, stdout, stderr)
        os.unlink(path)
        return

    msg = 'no secure delete programs were found'
    raise SecureDeleteError(msg, None, None)



# -*- coding: utf-8 -*-
"""
Drop account command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from __future__ import print_function

import sys

from clik import args, g, parser

from safe.cmd.drop import drop
from safe.cmd.list.account import print_prolixly
from safe.ec import CANCELED, NO_SUCH_ACCOUNT
from safe.model import Account
from safe.util import prompt_bool


@drop(alias='a')
def account():
    """Drop an account from the database."""
    parser.add_argument(
        'name',
        nargs=1,
        help='name/alias of account to drop',
    )

    yield

    account = Account.for_slug(args.name[0])
    if account is None:
        print('error: no account named', args.name[0], file=sys.stderr)
        yield NO_SUCH_ACCOUNT

    print()
    print_prolixly(account)
    print()
    if not prompt_bool('Delete this account?', default=False):
        yield CANCELED

    g.db.delete(account)
    g.commit_and_save()



# -*- coding: utf-8 -*-
"""
Policy form definitions.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from clik import g
from clik_wtforms import FieldList, Form, IntegerField, SelectField, \
    StringField
from wtforms.validators import InputRequired, NumberRange, Optional, \
    ValidationError

from safe.form import slug_validator
from safe.model import Policy
from safe.sgen import generate


class PolicyForm(Form):
    """Base form containing common fields, not meant to be used directly."""

    #: Short one-line description for the policy.
    #:
    #: :type: :class:`clik_wtforms.StringField`
    description = StringField(
        description='short one-line description for policy',
    )

    #: Frequency at which secret should be rotated.
    #:
    #: :type: :class:`clik_wtforms.IntegerField`
    frequency = IntegerField(
        description='number of days after which password should be rotated '
                    '(0 means never rotate) (default: '
                    '%s)' % Policy.DEFAULT_FREQUENCY,
        validators=[
            NumberRange(min=0, message='number must be 0 or greater'),
            Optional(),
        ],
    )

    #: Generator for secrets.
    #:
    #: :type: :class:`clik_wtforms.SelectField` of available
    #:        generators from :mod:`safe.sgen`
    generator = SelectField(
        choices=sorted(generate),
        description='method for auto-generating secrets for this policy '
                    '(default: %s)' % Policy.DEFAULT_GENERATOR,
    )

    #: Length of generated secrets.
    #:
    #: :type: :class:`clik_wtforms.IntegerField`
    length = IntegerField(
        description='length for auto-generated secrets (default: '
                    '%s)' % Policy.DEFAULT_LENGTH,
        validators=[
            NumberRange(min=1, message='number must be 1 or greater'),
            Optional(),
        ],
    )

    @staticmethod
    def get_short_arguments():
        """Return short arguments for the base fields."""
        return dict(
            d='description',
            f='frequency',
            g='generator',
            l='length',  # noqa: E741 (ambiguous var name makes sense here)
        )

    def update_policy(self, policy):
        """
        Modify ``policy`` based on the values from the form.

        :param safe.model.Policy policy: Policy to update
        """
        if self.description.data is not None:
            policy.description = self.description.data
        if self.frequency.data is not None:
            policy.frequency = self.frequency.data
        if self.generator.data is not None:
            policy.generator = self.generator.data
        if self.length.data is not None:
            policy.length = self.length.data


class NewPolicyForm(PolicyForm):
    """Form for creating new policies."""

    #: Characters that are not allowed in generated secrets.
    #:
    #: :type: :class:`clik_wtforms.FieldList` of
    #:        :class:`clik_wtforms.StringField`
    disallowed_characters = FieldList(
        StringField(),
        description='characters that may not be present in auto-generated '
                    'secrets',
    )

    #: Name for the policy.
    #:
    #: :type: :class:`clik_wtforms.StringField` validated by
    #:        :func:`safe.form.slug_validator`
    name = StringField(
        description='name for the policy',
        validators=[InputRequired(), slug_validator],
    )

    @classmethod
    def get_short_arguments(cls):
        """Return short arguments, merged with parent form short args."""
        d = super(NewPolicyForm, cls).get_short_arguments()
        d.update(dict(c='disallowed_characters', n='name'))
        return d

    def validate_name(self, field):
        """
        Validate that the new name does not already exist.

        :raises: :exc:`wtforms.validators.ValidationError` if new name is
                 the same as the name of an existing policy
        """
        if Policy.id_for_name(field.data) is not None:
            raise ValidationError('Policy with that name already exists')

    def create_policy(self):
        """
        Create a new policy based on the form data.

        Note that the new policy is added to the ``g.db`` session, but
        not committed.

        :return: Newly-created policy
        :rtype: :class:`safe.model.Policy`
        """
        policy = Policy(name=self.name.data)
        super(NewPolicyForm, self).update_policy(policy)
        chars = ''.join(sorted(set(''.join(self.disallowed_characters.data))))
        policy.disallowed_characters = chars
        g.db.add(policy)
        return policy


class UpdatePolicyForm(PolicyForm):
    """Form for updating an existing policy."""

    #: Character(s) to re-allow in generated secrets. This reverses
    #: the effect of :attr:`disallowed_characters`.
    #:
    #: :type: :class:`clik_wtforms.FieldList` of
    #:        :class:`clik_wtforms.StringField`
    allowed_characters = FieldList(
        StringField(),
        description='removes characters from the disallowed characters for '
                    'this policy (i.e. reverses the effect of '
                    '--disallowed-characters)',
    )

    #: Characters that are not allowed in generated secrets.
    #:
    #: :type: :class:`clik_wtforms.FieldList` of
    #:        :class:`clik_wtforms.StringField`
    disallowed_characters = FieldList(
        StringField(),
        description='characters that may not be part of the output',
    )

    #: New name for the policy.
    #:
    #: :type: :class:`clik_wtforms.StringField`
    new_name = StringField(
        description='new name for this policy (replaces current name)',
        metavar='NAME',
        validators=[slug_validator],
    )

    @classmethod
    def get_short_arguments(cls):
        """Return short arguments, merged with parent form short args."""
        d = super(UpdatePolicyForm, cls).get_short_arguments()
        d.update(dict(
            a='allowed_characters',
            c='disallowed_characters',
            n='new_name',
        ))
        return d

    def validate_new_name(self, field):
        """
        Validate new name.

        :param clik_wtforms.StringField field: New name field
        :raises: :exc:`wtforms.validators.ValidationError` if name is the same
                 as the current name, or already exists as a name of another
                 policy
        """
        field.change_name = False
        if field.data:
            if field.data == self.policy.name:
                msg = 'New name is the same as the current name'
                raise ValidationError(msg)
            if Policy.id_for_name(field.data) is not None:
                fmt = 'Policy with name "%s" already exists'
                raise ValidationError(fmt % field.data)
            field.change_name = True

    def bind_and_validate(self, policy, args=None):
        """Bind the ``policy`` to the form, then call the superclass."""
        self.policy = policy
        return super(UpdatePolicyForm, self).bind_and_validate(args)

    def update_policy(self):
        """
        Update the bound policy based on the data in the form.

        Note that this does not commit the changes.
        """
        super(UpdatePolicyForm, self).update_policy(self.policy)

        if self.new_name.change_name:
            self.policy.name = self.new_name.data

        allow, disallow = set(), set()
        for value in self.allowed_characters.data:
            for char in value:
                allow.add(char)
        for value in self.disallowed_characters.data:
            for char in value:
                disallow.add(char)
        chars = set(self.policy.disallowed_characters)
        for char in allow:
            if char in chars:
                chars.remove(char)
        for char in disallow:
            chars.add(char)
        self.policy.disallowed_characters = ''.join(sorted(chars))



# -*- coding: utf-8 -*-
"""
Package configuration for decafjoe-safe.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
import os

from setuptools import find_packages, setup


name = 'decafjoe-safe'
version = '0.7.0'
requires = (
    'clik==0.92.2',
    'clik-shell==0.90.0',
    'clik-wtforms==0.90.1',
    'sqlalchemy==1.2.8',
    'wtforms==2.2.1',
)

url = 'https://%s.readthedocs.io' % name
description = 'Password manager for people who like GPG and the command line.'
long_description = 'Please see the official project page at %s' % url

root_dir = os.path.abspath(os.path.dirname(__file__))
src_dir = os.path.join(root_dir, 'src')
packages = find_packages(src_dir)

package_name = 'safe'
entry_point = '%(name)s = %(name)s:main' % dict(name=package_name)

setup(
    author='Joe Joyce',
    author_email='joe@decafjoe.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    description=description,
    entry_points=dict(console_scripts=[entry_point]),
    install_requires=requires,
    license='BSD',
    long_description=long_description,
    name=name,
    package_dir={'': 'src'},
    packages=packages,
    url=url,
    version=version,
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""
New object command group.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2018.
:license: BSD
"""
from safe.app import safe


@safe
def new():
    """Add objects to the database."""
    yield
